/* rsadecrypt v1.0 Copyright (C) 2003 tomasliq <tom@eggdrop.ch>
 * 
 * Decrypts a file created with rsaencrypt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* includes */

/* ANSI C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* GMP library */
#include <gmp.h>

/* UNIX */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


/* defines */

#define VERSION "rsadecrypt v1.0 Copyright (C) 2003 tomasliq <tom@eggdrop.ch>"
#define NUMVER "010000"


/* prototypes */
int main(int argc, char *argv[]);
void usage(int argc, char *argv[]);
int file_exists(char *f);
void readkey(FILE *f, char *type, unsigned char **de, unsigned char **n);
void err(int num);
void crypt2plain(FILE *fc, FILE *fp, unsigned char **d, unsigned char **n);

/* functions */
int main(int argc, char *argv[])
{
	FILE *key, *plain, *crypt;
	unsigned char *d, *n;
	char type;
	
	/* print version */
	printf("%s\n", VERSION);

	/* check number of arguments */
	if (argc < 4) { usage(argc, argv); exit(1); }
	
	/* check to see whether the files exist or not */
	if (!file_exists(argv[1]))
	{
		printf("Error: Keyfile does not exist\n");
		exit(1);
	}
	
	if (!file_exists(argv[2]))
	{
		printf("Error: Cryptfile does not exist\n");
		exit(1);
	}

	if (file_exists(argv[3]))
	{
		printf("Error: Plainfile already exists\n");
		exit(1);
	}
	
	/* open the files */
	if ((key = fopen(argv[1], "r")) == NULL)
	{
		printf("Error: Can't open keyfile %s for reading\n", argv[1]);
		printf("Have you the required permissions?\n");
		exit(1);
	}

	if ((crypt = fopen(argv[2], "r")) == NULL)
	{
		printf("Error: Can't open cryptfile %s for reading\n", argv[2]);
		printf("Have you the required permissions?\n");
		exit(1);
	}

	if ((plain = fopen(argv[3], "w")) == NULL)
	{
		printf("Error: Can't open plainfile %s for writing\n", argv[1]);
		printf("Have you the required permissions?\n");
		exit(1);
	}
	
	/* read the key */
	readkey(key, &type, &d, &n);

	/* read plainfile */
	crypt2plain(crypt, plain, &d, &n);

	/* close the keyfile */
	fclose(key);

	/* print the key */

	free(d);
	free(n);

	/* and exit */
	return 0;
}

/* prints the usage */
void usage(int argc, char *argv[])
{
        printf("Usage: %s <privatekey> <cryptfile> <plainfile>\n", argv[0]);
}

/* returns 1 if a file exists or 0 if not*/
int file_exists(char *f) {
        struct stat s;
        return ((stat(f,&s)==0)?1:0);
}

void readkey(FILE *f, char *type, unsigned char **de, unsigned char **n)
{
	int c=0; /* stores the character read by fgetc */
	int i; /* used for loops */
	char numver[6];
	fpos_t filepos; /* stores the file position */
	long len; /* length of D/E or N */
	
	*type=0;

	/* the file starts with RSA */
	c = fgetc(f);
	if (c != 'R') err(1);
	c = fgetc(f);
	if (c != 'S') err(1);
	c = fgetc(f);
	if (c != 'A') err(1);

	/* read the numver */
	for (i=0;i<6;i++)
	{
		c = fgetc(f);
		if (c == EOF) err(1);
		if (!((c>='0')&&(c<='9'))) err(1);
		numver[i]=c;
	}

	if (!(numver[0] == '0' && numver[1] == '1')) err(3);

	/* read type of file */
	*type = fgetc(f);
	c = fgetc(f);
	if (c != ' ') err(1);
	if (*type != 'S') err(2);
	
	/* save the current position */
	if (fgetpos(f, &filepos) < 0)
	{
		printf("Error: fgetpos() failed\n");
		exit(1);
	}

	/* how long is D/E ? */
	i=0;
	while((c = fgetc(f)) != 32)
	{
		if (c < 32) err(1);
		i++;
	}
	len = i;

	/* now allocate memory */
	*de=(unsigned char *)malloc(len+1);
	if (*de == NULL) { printf("Error: No memory available\n"); exit(1); }

	c = fsetpos(f, &filepos);
	if (c == -1)
	{
		printf("Error: fsetpos() failed\n");
		exit(1);
	}

	/* write D/E into memory */
	i=0;
	while((c = fgetc(f)) != 32)
	{
		*(*de+i)=c;
		i++;
	}
	*(*de+i)='\0';

        /* save the current position */
        if (fgetpos(f, &filepos) < 0)
        {
                printf("Error: fgetpos() failed\n");
                exit(1);
        }

        /* how long is N ? */
        i=0;
        while((c = fgetc(f)) != '\n') /* last char is \n */
        {
                if (c <= 32) err(1);
                i++;
        }
        len = i;

        /* now allocate memory */
        *n=(unsigned char *)malloc(len+1);
	if (*n == NULL) { printf("Error: No memory available\n"); exit(1); }

        c = fsetpos(f, &filepos);
        if (c == -1)
        {
                printf("Error: fsetpos() failed\n");
                exit(1);
        }

        /* write N into memory */
        i=0;
        while((c = fgetc(f)) != '\n')
        {
                *(*n+i)=c;
                i++;
        }
        *(*n+i)='\0';

	return;
}

void err(int num)
{
	switch(num)
	{
		case 1:
			printf("Error: This is not a valid rsakeygen file\n");
			break;
		case 2:
			printf("Error: This is not a secret key file \n");
			break;
		case 3:
			printf("Error: Version should be 1.x.x\n");
			break;
	}
	exit(1);
}

void crypt2plain(FILE *fc, FILE *fp, unsigned char **d, unsigned char **n)
{
	unsigned char *tmpc;
	int i=0,c=0;
	int blocks=0, block=0, blocksize=0;
	long tmpfpos;
	char numver[6];
	fpos_t filepos;
	mpz_t nd, nn, base, tmp1, tmp2, crypted, num;

	mpz_init(nd); mpz_init(nn);
	mpz_init(base);
	mpz_init(tmp1); mpz_init(tmp2);
	mpz_init(crypted);
	mpz_init(num);

	mpz_set_ui(base, 223);

        mpz_set_ui(nd, *(*d)-33);
        for (i=1;i<strlen((char *)*d);i++)
        {       
                mpz_pow_ui(tmp1, base, i);
                mpz_mul_ui(tmp2, tmp1, *(*d+i)-33);
                mpz_set(tmp1, nd);
                mpz_add(nd, tmp1, tmp2);
        }

        mpz_set_ui(nn, *(*n)-33);
        for (i=1;i<strlen((char *)*n);i++)
        {       
                mpz_pow_ui(tmp1, base, i);
                mpz_mul_ui(tmp2, tmp1, *(*n+i)-33);
                mpz_set(tmp1, nn);
                mpz_add(nn, tmp1, tmp2);
        }	

	/* read file information, should be RSAnnnnnnC ... */
	c = fgetc(fc);
	if (c != 'R') err(1);
        c = fgetc(fc);
        if (c != 'S') err(1);
        c = fgetc(fc);
        if (c != 'A') err(1);
        /* read the numver */
        for (i=0;i<6;i++)
        {
                c = fgetc(fc);
                if (c == EOF) err(1);
                if (!((c>='0')&&(c<='9'))) err(1);
                numver[i]=c;
        }

	if (!(numver[0] == '0' && numver[1] == '1')) err(3);
	
	c = fgetc(fc);
	if (c != 'C') err(2);
	c = fgetc(fc);
	if (c != ' ') err(1);

	/* read number of blocks */
        if (fgetpos(fc, &filepos) < 0) 
        {                       
                printf("Error: fgetpos() failed\n");
                exit(1);
        }

	while((c = fgetc(fc)) != EOF)
	{
		if (blocks == 0)
		{
			blocksize++;
		}
		if (c == ' ') { blocks++; }
	}

	tmpfpos = ftell(fc);
	
        if (fsetpos(fc, &filepos) == -1)
        {
                printf("Error: fsetpos() failed\n");
                exit(1);
        }

	printf("Decrypting %d bytes (blocksize is %d bytes). This can take a while:\n", (int)tmpfpos, blocksize);

	printf("\r%10d of %10d blocks decrypted %3d%%", 0, blocks, 0);
	fflush(stdout);

	/* read the blocks */
	while(1)
	{
	        if (fgetpos(fc, &filepos) < 0)
	        {
	                printf("Error: fgetpos() failed\n");
	                exit(1);
	        }
		i=0;
	        while((c = fgetc(fc)) != 32)
	        {
			if (c == '\n') { goto done; } /* yeah i'm using goto */
	                if (c < 32) err(1);
	                i++;
	        }
		tmpc = (unsigned char *)malloc(i+1);
	        if (fsetpos(fc, &filepos) == -1)
	        {
	                printf("Error: fsetpos() failed\n");
	                exit(1);
	        }
		i=0;
	        while((c = fgetc(fc)) != 32)
	        {
			tmpc[i]=c;
			i++;
	        }
		tmpc[i]='\0';

		/* convert and decrypt tmpc */
		{
			int cnt;
			char *pl;
			
			mpz_set_ui(num, tmpc[0]-33);
			
			for (cnt=1;cnt<strlen((char *)tmpc);cnt++)
			{
				mpz_pow_ui(tmp1, base, cnt);
				mpz_mul_ui(tmp2, tmp1, tmpc[cnt]-33);
				mpz_set(tmp1, num);
				mpz_add(num, tmp1, tmp2);
			}
			/* printf("numZ = %s\n", mpz_get_str(NULL,10,num)); */
			mpz_powm(crypted, num, nd, nn);

			pl = malloc(strlen(mpz_get_str(NULL,17,crypted))+1);
			strcpy(pl, mpz_get_str(NULL,17,crypted));

			for (cnt=0;cnt<strlen(pl);cnt++)
			{
				if (pl[cnt] == 'a')
					pl[cnt]='9';
				else
					pl[cnt]-=1;
				if (cnt%2 == 1) 
				{
					int a=0,b=0;
					
					if (pl[cnt] >= '0' && pl[cnt] <= '9')
						a=pl[cnt]-'0';
					if (pl[cnt] >= 'a' && pl[cnt] <= 'f')
						a=10+pl[cnt]-'a';
					if (pl[cnt-1] >= '0' && pl[cnt-1] <= '9')
						b=pl[cnt-1]-'0';
					if (pl[cnt-1] >= 'a' && pl[cnt-1] <= 'f')
						b=10+pl[cnt-1]-'a';
					/* printf("%c*16+%c=\n", pl[cnt], pl[cnt-1]); */
					fputc(a*16+b, fp);
				}
			}
			block++;
			/* printf("pl = %d = %s\n", strlen(pl), pl); */
			printf("\r%10d of %10d blocks decrypted %3.0f%%", block, blocks, 100.0/(float)blocks*(float)block);
			fflush(stdout);
		}
	}
done:
	printf("\n");
	printf("Decrypting done!\n");

	return;
}
