/* rsakeygen v1.0 Copyright (C) 2003 tomasliq <tom@eggdrop.ch>
 * 
 * Generates RSA key
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* includes */

/* ANSI C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* GMP library */
#include <gmp.h>

/* UNIX */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/* own */
#include "primes.h"


/* defines */

#define RAND_ENGINE "/dev/urandom"
#define VERSION "rsakeygen v1.0 Copyright (C) 2003 tomasliq <tom@eggdrop.ch>"
#define NUMVER "010000"


/* prototypes */

int main(int argc, char *argv[]);
void usage(int argc, char *argv[]);
int file_exists(char *f);
void keygen(FILE *rand, FILE *private, FILE *public, int bits);
void gen_p(FILE *rand, mpz_t p, int bits);
void gen_q(FILE *rand, mpz_t p, mpz_t q, int bits);
void gen_d(FILE *rand, mpz_t d, mpz_t n, mpz_t q);
void calc_e(mpz_t e, mpz_t p, mpz_t q, mpz_t d);
int getprime(FILE *rand, mpz_t rnd, int bits);
void generate(FILE *rand, mpz_t ret, int bits);
void writekey(FILE *private, FILE *public, mpz_t n, mpz_t e, mpz_t d);

/* functions */

/* this is the main function */
int main(int argc, char *argv[])
{
	int bits;
	FILE *private, *public, *rand;

	/* print version */
	printf("%s\n", VERSION);
	
	/* check number of arguments */
	if (argc < 4) { usage(argc, argv); exit(1); }

	/* check bits */
	bits=atoi(argv[1]);
	if (bits == 0)
	{
		printf("Error: Bits must be a number\n");
		exit(1);
	}
	if ((bits%8) != 0)
	{
		printf("Error: Bits must be divisible by 8\n");
		exit(1);
	}
	if (bits<32) {
		printf("Error: Bits must be at least 32\n");
		exit(1);
	}

	/* check to see whether or not the public and
	 * private keyfiles already exist */
	if (file_exists(argv[2]))
	{
		printf("Error: File for privatekey already exists\n"); exit(1);
	}
	if (file_exists(argv[3]))
	{
		printf("Error: File for publickey already exists\n"); exit(1);
	}

	/* is privatekey file and publickey file the same? */
	if (!strcmp(argv[2], argv[3]))
	{
		printf("Error: File for privatekey is the same as file for \
publickey\n");
		exit(1);
	}

	/* open randengine, privatekey and publickey files */
	if ((rand = fopen(RAND_ENGINE, "r")) == NULL)
	{
		printf("Error: Can't open %s for reading\n", RAND_ENGINE);
		printf("Try to set RAND_ENGINE to your random engine and \
recompile rsakeygen.\n");
		exit(1);
	}
	if ((private = fopen(argv[2], "w")) == NULL)
	{
		printf("Error: Can't open %s for writing\n", argv[2]);
		printf("Have you the required permissions?\n");
		exit(1);
	}
	if ((public = fopen(argv[3], "w")) == NULL)
	{
		printf("Error: Can't open %s for writing\n", argv[3]);
		printf("Have you the required permissions?\n");
		exit(1);
	}

	/* now generate the key */
	keygen(rand, private, public, bits);

	/* close the files */
	fclose(rand);
	fclose(private);
	fclose(public);

	/* and exit */
	return 0;
}

/* prints the usage */
void usage(int argc, char *argv[])
{
        printf("Usage: %s <bits> <privatekey> <publickey>\n", argv[0]);
}

/* returns 1 if a file exists or 0 if not*/
int file_exists(char *f) {
	struct stat s;
	return ((stat(f,&s)==0)?1:0);
}

/* generates the keys */
void keygen(FILE *rand, FILE *private, FILE *public, int bits)
{
	mpz_t p, q, n, d, e;
	
	/* initialize variables */
	mpz_init(p); mpz_init(q); mpz_init(n); mpz_init(d); mpz_init(e);

	/* generate P */
	gen_p(rand, p, bits);

	/* generate Q */
	gen_q(rand, p, q, bits);

	/* q should be greater than p */
        if (mpz_cmp(p,q) > 0)
        {
                mpz_swap(p,q);
        }

	/* generate N */
	printf("Calculating N... ");
	fflush(stdout);
	/* n = p*q */
	mpz_mul(n, p, q);
	printf("OK\n");	

	/* generate D */
	gen_d(rand, d,n,q);

	/* calculate E */
	calc_e(e, p, q, d);

	/* now write the keyfiles */
	writekey(private, public, n, e, d);
}

/* generates P */
void gen_p(FILE *rand, mpz_t p, int bits)
{
	printf("Generating prime P.");
	fflush(stdout);
	for (;;)
	{
		/* generate prime */
		getprime(rand, p, bits);
		/* check prime */
		if (mpz_probab_prime_p(p, 5) != 0) { break; }
		printf(".");
		fflush(stdout);
		/* not a prime? then generate another prime */
	}
	printf("OK\n");
	return;
}

/* generates Q */
void gen_q(FILE *rand, mpz_t p, mpz_t q, int bits)
{
        printf("Generating prime Q.");
        fflush(stdout);
        for (;;)
        {
                /* generate prime */
                getprime(rand, q, bits);
                /* check prime */
		if (mpz_cmp(p,q) != 0) /* q cannot be p  */
		{
	                if (mpz_probab_prime_p(q, 5) != 0) { break; }
		}
                printf(".");
                fflush(stdout);
                /* not a prime? then generate another prime */
        }
        printf("OK\n");
        return;
}

/* generates D */
void gen_d(FILE *rand, mpz_t d, mpz_t n, mpz_t q)
{
	printf("Generating prime D.");
	fflush(stdout);
	for (;;)
	{
		/* getprime(rand, d,(strlen(mpz_get_str(NULL,2,n))+strlen(mpz_get_str(NULL,2,q)))/2); */
		getprime(rand, d, strlen(mpz_get_str(NULL,2,n)));
		if (mpz_cmp(d, q) > 0)
		{
			if (mpz_probab_prime_p (d, 5) != 0) { break; }
		}
		printf(".");
		fflush(stdout);
	}
	printf("OK\n");
}

/* calculates E */
void calc_e(mpz_t e, mpz_t p, mpz_t q, mpz_t d)
{
	mpz_t tmp_p, tmp_q, tmp_pq;

	/* initialize */
	mpz_init(tmp_p); mpz_init(tmp_q); mpz_init(tmp_pq);
	
	printf("Calculating E... ");
	fflush(stdout);
	
	/* calculate E */
	mpz_sub_ui(tmp_p, p, 1);
	mpz_sub_ui(tmp_q, q, 1);
	mpz_mul(tmp_pq, tmp_p, tmp_q);
	mpz_invert(e, d, tmp_pq);
	
	printf("OK\n");
}

/* generates a probable prime */
int getprime(FILE *rand, mpz_t rnd, int bits)
{
        int i;
        unsigned long mods[NUMPRIMES];
        unsigned long d, delta;
        unsigned long tmp_mod;
        mpz_t tmp;

	mpz_init(tmp);

loop1:
        generate(rand, rnd, bits);
        tmp_mod=mpz_mod_ui(tmp, rnd, 2);
        if (tmp_mod==0) { goto loop1; } /* stupid goto :) */

        for (i=1; i<NUMPRIMES; i++)
        {
                mods[i]=mpz_mod_ui(tmp, rnd, primes[i]);
        }
        delta=0;

loop2:
        for (i=1; i<NUMPRIMES; i++)
        {
                if ( ((mods[i]+delta)%primes[i]) <= 1)
                {
                        d=delta;
                        delta+=2;
                        if (delta < d) goto loop1;
                        goto loop2; /* stupid goto again :) */
                }
        }
        mpz_add_ui(tmp,rnd,delta);
                
        mpz_set(rnd,tmp);
                        
                        
        for (i=0; i<NUMPRIMES; i++)
        {       
                tmp_mod=mpz_mod_ui(tmp, rnd, primes[i]);
                if (tmp_mod == 0) { goto loop1; } /* ... */
        }       
	return 0;
}               

/* returns a random number */
void generate(FILE *rand, mpz_t ret, int bits)
{
        unsigned char c;
        char *buf;
        char *str_num;
        int i, hex_i;
        int bytes;

        bytes=(bits+7)/8;

        /* allocate some needed memory */
        buf = (char *)malloc(bytes); /* we don't need +1 for '\0' */
        if (buf == NULL) { printf("Error: malloc failed!\n"); exit(1); }
        str_num = (char *)malloc((bytes*2)+1); /* +1 for '\0' */
        if (str_num == NULL) { printf("Error: malloc failed!\n"); exit(1); }

        /* read */
        fread(buf, bytes, 1, rand);

        hex_i=0;

        for (i=0;i<bytes;i++)
        {
                int hex;
                char hx;
                c = buf[i];

                hex=0;
                if (c & 1) hex+=1;
                if (c & 2) hex+=2;
                if (c & 4) hex+=4;
                if (c & 8) hex+=8;
                if (hex <=  9) { hx = '0' + hex; }
                if (hex >= 10) { hx = 'A' + (hex-10); }

                /* append to buffer */
                str_num[hex_i]=hx;
                hex_i++;
                hex=0;
                if (c &  16) hex+=1;
                if (c &  32) hex+=2;
                if (c &  64) hex+=4;
                if (c & 128) hex+=8;
                if (hex <=  9) { hx = '0' + hex; }
                if (hex >= 10) { hx = 'A' + (hex-10); }

                /* append to buffer */
                str_num[hex_i]=hx;
                hex_i++;
        }
        str_num[hex_i]='\0';

        /* set number ret with base 16 (hexadecimal) */
        mpz_set_str(ret, str_num, 16);

        free(buf);
        free(str_num);

        return;
}

/* writes the key */
void writekey(FILE *private, FILE *public, mpz_t n, mpz_t e, mpz_t d)
{
	int base;
	mpz_t tmp_num, rest, tmp;

	mpz_init(tmp_num); mpz_init(rest); mpz_init(tmp);

	/* syntax of a private key file:
	 * RSA<numver>S <D> <N>\n
	 *
	 * syntax of a public key file:
	 * RSA<numver>P <E> <N>\n
	 *
	 * <numver> has exatly six bytes (#define NUMVER)
	 */

	printf("Writing key... ");
	fflush(stdout);

	/* put rsatools info into the files */
	fprintf(private, "RSA%sS ", NUMVER);
	fprintf(public, "RSA%sP ", NUMVER);

	/* convert it into one byte blocks:
	 * base 256-33=223; we don't use the first 32 chars */
	base=223;

        /* put D into privatekey file */
        mpz_set(tmp_num,d);
        while(mpz_cmp_ui(tmp_num,0) != 0)
        {
                mpz_mod_ui(rest, tmp_num, base);
                fputc(atoi(mpz_get_str(NULL,0,rest))+33, private);
                mpz_sub(tmp, tmp_num, rest);
                mpz_tdiv_q_ui(tmp_num, tmp, base);
        }
        fputc(' ', private);

	/* put E into publickey file */
	mpz_set(tmp_num,e);
	while(mpz_cmp_ui(tmp_num,0) != 0)
	{
		mpz_mod_ui(rest, tmp_num, base);
		fputc(atoi(mpz_get_str(NULL,0,rest))+33, public);
		mpz_sub(tmp, tmp_num, rest);
		mpz_tdiv_q_ui(tmp_num, tmp, base);
	}
	fputc(' ', public);
	
	/* put N into publickey and privatekey file */
	mpz_set(tmp_num,n);
        while(mpz_cmp_ui(tmp_num,0) != 0)
        {
                mpz_mod_ui(rest, tmp_num, base);
                fputc(atoi(mpz_get_str(NULL,0,rest))+33, private);
		fputc(atoi(mpz_get_str(NULL,0,rest))+33, public);
                mpz_sub(tmp, tmp_num, rest);
                mpz_tdiv_q_ui(tmp_num, tmp, base);
        }
	
	/* finally put a line feed into the files */
	fputc('\n', private);
	fputc('\n', public);

	printf("OK\n");
}

