/* rsareadkey v1.0 Copyright (C) 2003 tomasliq <tom@eggdrop.ch>
 * 
 * Reads the RSA key from a file crated with rsakeygen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* includes */

/* ANSI C */
#include <stdio.h>
#include <stdlib.h>

/* GMP library */
#include <gmp.h>

/* UNIX */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


/* defines */

#define VERSION "rsareadkey v1.0 Copyright (C) 2003 tomasliq <tom@eggdrop.ch>"
#define NUMVER "010000"


/* prototypes */
int main(int argc, char *argv[]);
void usage(int argc, char *argv[]);
int file_exists(char *f);
void readkey(FILE *f, char *type, unsigned char **de, unsigned char **n, unsigned char *numver);
void printkey(char *name, char type, unsigned char **de, unsigned char **n, unsigned char *numver);
void err(int num);

/* functions */
int main(int argc, char *argv[])
{
	FILE *key;
	unsigned char *de, *n;
	char type;
	unsigned char numver[6];
	
	/* print version */
	printf("%s\n", VERSION);

	/* check number of arguments */
	if (argc < 2) { usage(argc, argv); exit(1); }
	
	/* check to see whether the keyfile exist */
	if (!file_exists(argv[1]))
	{
		printf("Error: Keyfile does not exist\n");
		exit(1);
	}
	
	/* open the file */
	if ((key = fopen(argv[1], "r")) == NULL)
	{
		printf("Error: Can't open keyfile %s for reading\n", argv[1]);
		printf("Have you the required permissions?\n");
		exit(1);
	}
	
	/* read the key */
	readkey(key, &type, &de, &n, numver);

	/* close the keyfile */
	fclose(key);

	/* print the key */
	printkey(argv[1], type, &de, &n, numver);

	free(de);
	free(n);

	/* and exit */
	return 0;
}

/* prints the usage */
void usage(int argc, char *argv[])
{
        printf("Usage: %s <keyfile>\n", argv[0]);
}

/* returns 1 if a file exists or 0 if not*/
int file_exists(char *f) {
        struct stat s;
        return ((stat(f,&s)==0)?1:0);
}

void readkey(FILE *f, char *type, unsigned char **de, unsigned char **n, unsigned char *numver)
{
	int c=0; /* stores the character read by fgetc */
	int i; /* used for loops */
	/* char numver[6]; */
	fpos_t filepos; /* stores the file position */
	long len; /* length of D/E or N */
	
	*type=0;

	/* the file starts with RSA */
	c = fgetc(f);
	if (c != 'R') err(1);
	c = fgetc(f);
	if (c != 'S') err(1);
	c = fgetc(f);
	if (c != 'A') err(1);

	/* read the numver */
	for (i=0;i<6;i++)
	{
		c = fgetc(f);
		if (c == EOF) err(1);
		if (!((c>='0')&&(c<='9'))) err(1);
		numver[i]=c;
	}

	/* read type of file */
	*type = fgetc(f);
	c = fgetc(f);
	if (c != ' ') err(1);
	if ((*type != 'S') && (*type != 'P')) err(2);
	
	/* save the current position */
	if (fgetpos(f, &filepos) < 0)
	{
		printf("Error: fgetpos() failed\n");
		exit(1);
	}

	/* how long is D/E ? */
	i=0;
	while((c = fgetc(f)) != 32)
	{
		if (c < 32) err(1);
		i++;
	}
	len = i;

	/* now allocate memory */
	*de=(unsigned char *)malloc(len+1);
	if (*de == NULL) { printf("Error: No memory available\n"); exit(1); }

	c = fsetpos(f, &filepos);
	if (c == -1)
	{
		printf("Error: fsetpos() failed\n");
		exit(1);
	}

	/* write D/E into memory */
	i=0;
	while((c = fgetc(f)) != 32)
	{
		*(*de+i)=c;
		i++;
	}
	*(*de+i)='\0';

        /* save the current position */
        if (fgetpos(f, &filepos) < 0)
        {
                printf("Error: fgetpos() failed\n");
                exit(1);
        }

        /* how long is N ? */
        i=0;
        while((c = fgetc(f)) != '\n') /* last char is \n */
        {
                if (c <= 32) err(1);
                i++;
        }
        len = i;

        /* now allocate memory */
        *n=(unsigned char *)malloc(len+1);
	if (*n == NULL) { printf("Error: No memory available\n"); exit(1); }

        c = fsetpos(f, &filepos);
        if (c == -1)
        {
                printf("Error: fsetpos() failed\n");
                exit(1);
        }

        /* write N into memory */
        i=0;
        while((c = fgetc(f)) != '\n')
        {
                *(*n+i)=c;
                i++;
        }
        *(*n+i)='\0';

	return;
}

void err(int num)
{
	switch(num)
	{
		case 1:
			printf("Error: This is not a valid rsakeygen file\n");
			break;
		case 2:
			printf("Error: This is not a private or public key \
file \n");
			break;
	}
	exit(1);
}
void printkey(char *name, char type, unsigned char **de, unsigned char **n, unsigned char *numver)
{
	mpz_t num, tmp1, tmp2, base;
	int i;

	mpz_init(num); mpz_init(tmp1); mpz_init(tmp2); mpz_init(base);
	
	printf(" Filename: %s\n", name);
	printf(" Version:  ");
	if (numver[0]!='0') { putchar(numver[0]); }
	putchar(numver[1]); putchar('.');
	if (numver[2]!='0') { putchar(numver[2]); }
	putchar(numver[3]); putchar('.');
	if (numver[4]!='0') { putchar(numver[4]); }
	putchar(numver[5]); putchar('\n');
	printf(" Type:     %s\n", type == 'P'?"publickey":"privatekey");

	/* base is 223 */
	mpz_set_ui(base, 223);

	mpz_set_ui(num, *(*de)-33);
	
	for (i=1;i<strlen((char *)*de);i++)
	{
		mpz_pow_ui(tmp1, base, i);
		mpz_mul_ui(tmp2, tmp1, *(*de+i)-33);
		mpz_set(tmp1, num);
		mpz_add(num, tmp1, tmp2);
	}
	
	printf(" %c:\n%s\n", type == 'P'?'E':'D', mpz_get_str(NULL, 0, num));

	mpz_set_ui(num, *(*n)-33);

	for (i=1;i<strlen((char *)*n);i++)
	{
		mpz_pow_ui(tmp1, base, i);
		mpz_mul_ui(tmp2, tmp1, *(*n+i)-33); 
		mpz_set(tmp1, num);
		mpz_add(num, tmp1, tmp2);
	}
	
	printf(" N:\n%s\n", mpz_get_str(NULL, 0, num));
}

